import nerodia


class Container(object):
    def element(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, self._extract_selector(*args, **kwargs))

    def elements(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, self._extract_selector(*args, **kwargs))

    # private

    def _extract_selector(self, *args, **kwargs):
        if args and len(args) == 2:
            nerodia.logger.deprecate('Using ordered parameters to locate elements '
                                     '({}, {})'.format(*args), '{{{}={}}}'.format(*args),
                                     ids='selector_parameters')
            return {args[0]: args[1]}
        elif not args:
            return kwargs

        raise ValueError('expected kwargs dict, got {}'.format(kwargs))

    # Plural of 'a' cannot be a method name, use link/links instead
    def link(self, *args, **kwargs):
        from .elements.link import Anchor
        return Anchor(self, dict(self._extract_selector(*args, **kwargs), tag_name='a'))

    def links(self, *args, **kwargs):
        from .elements.html_elements import AnchorCollection
        return AnchorCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='a'))

    def abbr(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='abbr'))

    def abbrs(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                     tag_name='abbr'))

    def address(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='address'))

    def addresses(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                     tag_name='address'))

    def area(self, *args, **kwargs):
        from .elements.area import Area
        return Area(self, dict(self._extract_selector(*args, **kwargs), tag_name='area'))

    def areas(self, *args, **kwargs):
        from .elements.html_elements import AreaCollection
        return AreaCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='area'))

    def article(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='article'))

    def articles(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='article'))

    def aside(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='aside'))

    def asides(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='aside'))

    def audio(self, *args, **kwargs):
        from .elements.html_elements import Audio
        return Audio(self, dict(self._extract_selector(*args, **kwargs), tag_name='audio'))

    def audios(self, *args, **kwargs):
        from .elements.html_elements import AudioCollection
        return AudioCollection(self, dict(self._extract_selector(*args, **kwargs),
                                          tag_name='audio'))

    def b(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='b'))

    def bs(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='b'))

    def base(self, *args, **kwargs):
        from .elements.html_elements import Base
        return Base(self, dict(self._extract_selector(*args, **kwargs), tag_name='base'))

    def bases(self, *args, **kwargs):
        from .elements.html_elements import BaseCollection
        return BaseCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='base'))

    def bdi(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='bdi'))

    def bdis(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='bdi'))

    def bdo(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='bdo'))

    def bdos(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='bdo'))

    def blockquote(self, *args, **kwargs):
        from .elements.html_elements import Quote
        return Quote(self, dict(self._extract_selector(*args, **kwargs), tag_name='blockquote'))

    def blockquotes(self, *args, **kwargs):
        from .elements.html_elements import QuoteCollection
        return QuoteCollection(self, dict(self._extract_selector(*args, **kwargs),
                                          tag_name='blockquote'))

    def body(self, *args, **kwargs):
        from .elements.html_elements import Body
        return Body(self, dict(self._extract_selector(*args, **kwargs), tag_name='body'))

    def bodys(self, *args, **kwargs):
        from .elements.html_elements import BodyCollection
        return BodyCollection(self, dict(self._extract_selector(*args, **kwargs),
                                         tag_name='body'))

    def br(self, *args, **kwargs):
        from .elements.html_elements import BR
        return BR(self, dict(self._extract_selector(*args, **kwargs), tag_name='br'))

    def brs(self, *args, **kwargs):
        from .elements.html_elements import BRCollection
        return BRCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='br'))

    def button(self, *args, **kwargs):
        from .elements.button import Button
        return Button(self, dict(self._extract_selector(*args, **kwargs), tag_name='button'))

    def buttons(self, *args, **kwargs):
        from .elements.html_elements import ButtonCollection
        return ButtonCollection(self, dict(self._extract_selector(*args, **kwargs),
                                           tag_name='button'))

    def canvas(self, *args, **kwargs):
        from .elements.html_elements import Canvas
        return Canvas(self, dict(self._extract_selector(*args, **kwargs), tag_name='canvas'))

    def canvases(self, *args, **kwargs):
        from .elements.html_elements import CanvasCollection
        return CanvasCollection(self, dict(self._extract_selector(*args, **kwargs),
                                           tag_name='canvas'))

    def caption(self, *args, **kwargs):
        from .elements.html_elements import TableCaption
        return TableCaption(self, dict(self._extract_selector(*args, **kwargs),
                                       tag_name='caption'))

    def captions(self, *args, **kwargs):
        from .elements.html_elements import TableCaptionCollection
        return TableCaptionCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                 tag_name='caption'))

    def checkbox(self, *args, **kwargs):
        from .elements.check_box import CheckBox
        return CheckBox(self, dict(self._extract_selector(*args, **kwargs), tag_name='input',
                                   type='checkbox',))

    def checkboxes(self, *args, **kwargs):
        from .elements.check_box import CheckBoxCollection
        return CheckBoxCollection(self, dict(self._extract_selector(*args, **kwargs),
                                             tag_name='input', type='checkbox',))

    def cite(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='cite'))

    def cites(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='cite'))

    def code(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='code'))

    def codes(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='code'))

    def col(self, *args, **kwargs):
        from .elements.html_elements import TableCol
        return TableCol(self, dict(self._extract_selector(*args, **kwargs), tag_name='col'))

    def cols(self, *args, **kwargs):
        from .elements.html_elements import TableColCollection
        return TableColCollection(self, dict(self._extract_selector(*args, **kwargs),
                                             tag_name='col'))

    def colgroup(self, *args, **kwargs):
        from .elements.html_elements import TableCol
        return TableCol(self, dict(self._extract_selector(*args, **kwargs), tag_name='colgroup'))

    def colgroups(self, *args, **kwargs):
        from .elements.html_elements import TableColCollection
        return TableColCollection(self, dict(self._extract_selector(*args, **kwargs),
                                             tag_name='colgroup'))

    def data(self, *args, **kwargs):
        from .elements.html_elements import Data
        return Data(self, dict(self._extract_selector(*args, **kwargs), tag_name='data'))

    def datas(self, *args, **kwargs):
        from .elements.html_elements import DataCollection
        return DataCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='data'))

    def datalist(self, *args, **kwargs):
        from .elements.html_elements import DataList
        return DataList(self, dict(self._extract_selector(*args, **kwargs), tag_name='datalist'))

    def datalists(self, *args, **kwargs):
        from .elements.html_elements import DataListCollection
        return DataListCollection(self, dict(self._extract_selector(*args, **kwargs),
                                             tag_name='datalist'))

    def date_field(self, *args, **kwargs):
        from .elements.date_field import DateField
        return DateField(self, dict(self._extract_selector(*args, **kwargs), tag_name='input',
                                    type='date'))

    def date_fields(self, *args, **kwargs):
        from .elements.date_field import DateFieldCollection
        return DateFieldCollection(self, dict(self._extract_selector(*args, **kwargs),
                                              tag_name='input', type='date'))

    def date_time_field(self, *args, **kwargs):
        from .elements.date_time_field import DateTimeField
        return DateTimeField(self, dict(self._extract_selector(*args, **kwargs), tag_name='input',
                                        type='datetime-local'))

    def date_time_fields(self, *args, **kwargs):
        from .elements.date_time_field import DateTimeFieldCollection
        return DateTimeFieldCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                  tag_name='input', type='datetime-local'))

    def dd(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='dd'))

    def dds(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='dd'))

    # 'del' is an invalid method name, use delete/deletes instead
    def delete(self, *args, **kwargs):
        from .elements.html_elements import Mod
        return Mod(self, dict(self._extract_selector(*args, **kwargs), tag_name='del'))

    def deletes(self, *args, **kwargs):
        from .elements.html_elements import ModCollection
        return ModCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='del'))

    def details(self, *args, **kwargs):
        from .elements.html_elements import Details
        return Details(self, dict(self._extract_selector(*args, **kwargs), tag_name='details'))

    def detailses(self, *args, **kwargs):
        from .elements.html_elements import DetailsCollection
        return DetailsCollection(self, dict(self._extract_selector(*args, **kwargs),
                                            tag_name='details'))

    def dfn(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='dfn'))

    def dfns(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='dfn'))

    def dialog(self, *args, **kwargs):
        from .elements.html_elements import Dialog
        return Dialog(self, dict(self._extract_selector(*args, **kwargs), tag_name='dialog'))

    def dialogs(self, *args, **kwargs):
        from .elements.html_elements import DialogCollection
        return DialogCollection(self, dict(self._extract_selector(*args, **kwargs),
                                           tag_name='dialog'))

    def div(self, *args, **kwargs):
        from .elements.html_elements import Div
        return Div(self, dict(self._extract_selector(*args, **kwargs), tag_name='div'))

    def divs(self, *args, **kwargs):
        from .elements.html_elements import DivCollection
        return DivCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='div'))

    def dl(self, *args, **kwargs):
        from .elements.d_list import DList
        return DList(self, dict(self._extract_selector(*args, **kwargs), tag_name='dl'))

    def dls(self, *args, **kwargs):
        from .elements.html_elements import DListCollection
        return DListCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='dl'))

    def dt(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='dt'))

    def dts(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='dt'))

    def em(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='em'))

    def ems(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='em'))

    def embed(self, *args, **kwargs):
        from .elements.html_elements import Embed
        return Embed(self, dict(self._extract_selector(*args, **kwargs), tag_name='embed'))

    def embeds(self, *args, **kwargs):
        from .elements.html_elements import EmbedCollection
        return EmbedCollection(self, dict(self._extract_selector(*args, **kwargs),
                                          tag_name='embed'))

    def fieldset(self, *args, **kwargs):
        from .elements.html_elements import FieldSet
        return FieldSet(self, dict(self._extract_selector(*args, **kwargs), tag_name='fieldset'))

    field_set = fieldset

    def fieldsets(self, *args, **kwargs):
        from .elements.html_elements import FieldSetCollection
        return FieldSetCollection(self, dict(self._extract_selector(*args, **kwargs),
                                             tag_name='fieldset'))

    field_sets = fieldsets

    def figcaption(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs),
                                      tag_name='figcaption'))

    def figcaptions(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='figcaption'))

    def figure(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='figure'))

    def figures(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='figure'))

    def file_field(self, *args, **kwargs):
        from .elements.file_field import FileField
        return FileField(self, dict(self._extract_selector(*args, **kwargs), tag_name='input',
                                    type='file'))

    def file_fields(self, *args, **kwargs):
        from .elements.file_field import FileFieldCollection
        return FileFieldCollection(self, dict(self._extract_selector(*args, **kwargs),
                                              tag_name='input', type='file'))

    def font(self, *args, **kwargs):
        from .elements.font import Font
        return Font(self, dict(self._extract_selector(*args, **kwargs), tag_name='font'))

    def fonts(self, *args, **kwargs):
        from .elements.html_elements import FontCollection
        return FontCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='font'))

    def footer(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='footer'))

    def footers(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='footer'))

    def form(self, *args, **kwargs):
        from .elements.form import Form
        return Form(self, dict(self._extract_selector(*args, **kwargs), tag_name='form'))

    def forms(self, *args, **kwargs):
        from .elements.html_elements import FormCollection
        return FormCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='form'))

    def frame(self, *args, **kwargs):
        from .elements.i_frame import Frame
        return Frame(self, dict(self._extract_selector(*args, **kwargs), tag_name='frame'))

    def frames(self, *args, **kwargs):
        from .elements.i_frame import FrameCollection
        return FrameCollection(self, dict(self._extract_selector(*args, **kwargs),
                                          tag_name='frame'))

    def frameset(self, *args, **kwargs):
        from .elements.html_elements import FrameSet
        return FrameSet(self, dict(self._extract_selector(*args, **kwargs), tag_name='frameset'))

    def framesets(self, *args, **kwargs):
        from .elements.html_elements import FrameSetCollection
        return FrameSetCollection(self, dict(self._extract_selector(*args, **kwargs),
                                             tag_name='frameset'))

    def h1(self, *args, **kwargs):
        from .elements.html_elements import Heading
        return Heading(self, dict(self._extract_selector(*args, **kwargs), tag_name='h1'))

    def h1s(self, *args, **kwargs):
        from .elements.html_elements import HeadingCollection
        return HeadingCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='h1'))

    def h2(self, *args, **kwargs):
        from .elements.html_elements import Heading
        return Heading(self, dict(self._extract_selector(*args, **kwargs), tag_name='h2'))

    def h2s(self, *args, **kwargs):
        from .elements.html_elements import HeadingCollection
        return HeadingCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='h2'))

    def h3(self, *args, **kwargs):
        from .elements.html_elements import Heading
        return Heading(self, dict(self._extract_selector(*args, **kwargs), tag_name='h3'))

    def h3s(self, *args, **kwargs):
        from .elements.html_elements import HeadingCollection
        return HeadingCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='h3'))

    def h4(self, *args, **kwargs):
        from .elements.html_elements import Heading
        return Heading(self, dict(self._extract_selector(*args, **kwargs), tag_name='h4'))

    def h4s(self, *args, **kwargs):
        from .elements.html_elements import HeadingCollection
        return HeadingCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='h4'))

    def h5(self, *args, **kwargs):
        from .elements.html_elements import Heading
        return Heading(self, dict(self._extract_selector(*args, **kwargs), tag_name='h5'))

    def h5s(self, *args, **kwargs):
        from .elements.html_elements import HeadingCollection
        return HeadingCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='h5'))

    def h6(self, *args, **kwargs):
        from .elements.html_elements import Heading
        return Heading(self, dict(self._extract_selector(*args, **kwargs), tag_name='h6'))

    def h6s(self, *args, **kwargs):
        from .elements.html_elements import HeadingCollection
        return HeadingCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='h6'))

    def head(self, *args, **kwargs):
        from .elements.html_elements import Head
        return Head(self, dict(self._extract_selector(*args, **kwargs), tag_name='head'))

    def heads(self, *args, **kwargs):
        from .elements.html_elements import HeadCollection
        return HeadCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='head'))

    def header(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='header'))

    def headers(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='header'))

    def hgroup(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='hgroup'))

    def hgroups(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='hgroup'))

    def hidden(self, *args, **kwargs):
        from .elements.hidden import Hidden
        return Hidden(self, dict(self._extract_selector(*args, **kwargs), tag_name='input',
                                 type='hidden'))

    def hiddens(self, *args, **kwargs):
        from .elements.hidden import HiddenCollection
        return HiddenCollection(self, dict(self._extract_selector(*args, **kwargs),
                                           tag_name='input', type='hidden'))

    def hr(self, *args, **kwargs):
        from .elements.html_elements import HR
        return HR(self, dict(self._extract_selector(*args, **kwargs), tag_name='hr'))

    def hrs(self, *args, **kwargs):
        from .elements.html_elements import HRCollection
        return HRCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='hr'))

    def html(self, *args, **kwargs):
        from .elements.html_elements import Html
        return Html(self, dict(self._extract_selector(*args, **kwargs), tag_name='html'))

    def htmls(self, *args, **kwargs):
        from .elements.html_elements import HtmlCollection
        return HtmlCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='html'))

    # Plural of 'i' cannot be a method name, use ital/itals instead
    def ital(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='i'))

    def itals(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='i'))

    def iframe(self, *args, **kwargs):
        from .elements.i_frame import IFrame
        return IFrame(self, dict(self._extract_selector(*args, **kwargs), tag_name='iframe'))

    def iframes(self, *args, **kwargs):
        from .elements.html_elements import IFrameCollection
        return IFrameCollection(self, dict(self._extract_selector(*args, **kwargs),
                                           tag_name='iframe'))

    def img(self, *args, **kwargs):
        from .elements.image import Image
        return Image(self, dict(self._extract_selector(*args, **kwargs), tag_name='img'))

    image = img

    def imgs(self, *args, **kwargs):
        from .elements.html_elements import ImageCollection
        return ImageCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='img'))

    images = imgs

    def input(self, *args, **kwargs):
        from .elements.input import Input
        return Input(self, dict(self._extract_selector(*args, **kwargs), tag_name='input'))

    def inputs(self, *args, **kwargs):
        from .elements.html_elements import InputCollection
        return InputCollection(self, dict(self._extract_selector(*args, **kwargs),
                                          tag_name='input'))

    def ins(self, *args, **kwargs):
        from .elements.html_elements import Mod
        return Mod(self, dict(self._extract_selector(*args, **kwargs), tag_name='ins'))

    def inses(self, *args, **kwargs):
        from .elements.html_elements import ModCollection
        return ModCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='ins'))

    def kbd(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='kbd'))

    def kbds(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='kbd'))

    def keygen(self, *args, **kwargs):
        from .elements.html_elements import Keygen
        return Keygen(self, dict(self._extract_selector(*args, **kwargs), tag_name='keygen'))

    def keygens(self, *args, **kwargs):
        from .elements.html_elements import KeygenCollection
        return KeygenCollection(self, dict(self._extract_selector(*args, **kwargs),
                                           tag_name='keygen'))

    def label(self, *args, **kwargs):
        from .elements.html_elements import Label
        return Label(self, dict(self._extract_selector(*args, **kwargs), tag_name='label'))

    def labels(self, *args, **kwargs):
        from .elements.html_elements import LabelCollection
        return LabelCollection(self, dict(self._extract_selector(*args, **kwargs),
                                          tag_name='label'))

    def legend(self, *args, **kwargs):
        from .elements.html_elements import Legend
        return Legend(self, dict(self._extract_selector(*args, **kwargs), tag_name='legend'))

    def legends(self, *args, **kwargs):
        from .elements.html_elements import LegendCollection
        return LegendCollection(self, dict(self._extract_selector(*args, **kwargs),
                                           tag_name='legend'))

    def li(self, *args, **kwargs):
        from .elements.html_elements import LI
        return LI(self, dict(self._extract_selector(*args, **kwargs), tag_name='li'))

    def lis(self, *args, **kwargs):
        from .elements.html_elements import LICollection
        return LICollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='li'))

    def main(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='main'))

    def mains(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='main'))

    def map(self, *args, **kwargs):
        from .elements.html_elements import Map
        return Map(self, dict(self._extract_selector(*args, **kwargs), tag_name='map'))

    def maps(self, *args, **kwargs):
        from .elements.html_elements import MapCollection
        return MapCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='map'))

    def mark(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='mark'))

    def marks(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='mark'))

    def menu(self, *args, **kwargs):
        from .elements.html_elements import Menu
        return Menu(self, dict(self._extract_selector(*args, **kwargs), tag_name='menu'))

    def menus(self, *args, **kwargs):
        from .elements.html_elements import MenuCollection
        return MenuCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='menu'))

    def menuitem(self, *args, **kwargs):
        from .elements.html_elements import MenuItem
        return MenuItem(self, dict(self._extract_selector(*args, **kwargs), tag_name='menuitem'))

    def menuitems(self, *args, **kwargs):
        from .elements.html_elements import MenuItemCollection
        return MenuItemCollection(self, dict(self._extract_selector(*args, **kwargs),
                                             tag_name='menuitem'))

    def meta(self, *args, **kwargs):
        from .elements.html_elements import Meta
        return Meta(self, dict(self._extract_selector(*args, **kwargs), tag_name='meta'))

    def metas(self, *args, **kwargs):
        from .elements.html_elements import MetaCollection
        return MetaCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='meta'))

    def meter(self, *args, **kwargs):
        from .elements.html_elements import Meter
        return Meter(self, dict(self._extract_selector(*args, **kwargs), tag_name='meter'))

    def meters(self, *args, **kwargs):
        from .elements.html_elements import MeterCollection
        return MeterCollection(self, dict(self._extract_selector(*args, **kwargs),
                                          tag_name='meter'))

    def nav(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='nav'))

    def navs(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='nav'))

    def noscript(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='noscript'))

    def noscripts(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='noscript'))

    def object(self, *args, **kwargs):
        from .elements.html_elements import Object
        return Object(self, dict(self._extract_selector(*args, **kwargs), tag_name='object'))

    def objects(self, *args, **kwargs):
        from .elements.html_elements import ObjectCollection
        return ObjectCollection(self, dict(self._extract_selector(*args, **kwargs),
                                           tag_name='object'))

    def ol(self, *args, **kwargs):
        from .elements.list import OList
        return OList(self, dict(self._extract_selector(*args, **kwargs), tag_name='ol'))

    def ols(self, *args, **kwargs):
        from .elements.html_elements import OListCollection
        return OListCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='ol'))

    def optgroup(self, *args, **kwargs):
        from .elements.html_elements import OptGroup
        return OptGroup(self, dict(self._extract_selector(*args, **kwargs), tag_name='optgroup'))

    def optgroups(self, *args, **kwargs):
        from .elements.html_elements import OptGroupCollection
        return OptGroupCollection(self, dict(self._extract_selector(*args, **kwargs),
                                             tag_name='optgroup'))

    def option(self, *args, **kwargs):
        from .elements.option import Option
        return Option(self, dict(self._extract_selector(*args, **kwargs), tag_name='option'))

    def options(self, *args, **kwargs):
        from .elements.html_elements import OptionCollection
        return OptionCollection(self, dict(self._extract_selector(*args, **kwargs),
                                           tag_name='option'))

    def output(self, *args, **kwargs):
        from .elements.html_elements import Output
        return Output(self, dict(self._extract_selector(*args, **kwargs), tag_name='output'))

    def outputs(self, *args, **kwargs):
        from .elements.html_elements import OutputCollection
        return OutputCollection(self, dict(self._extract_selector(*args, **kwargs),
                                           tag_name='output'))

    def p(self, *args, **kwargs):
        from .elements.html_elements import Paragraph
        return Paragraph(self, dict(self._extract_selector(*args, **kwargs), tag_name='p'))

    def ps(self, *args, **kwargs):
        from .elements.html_elements import ParagraphCollection
        return ParagraphCollection(self, dict(self._extract_selector(*args, **kwargs),
                                              tag_name='p'))

    def param(self, *args, **kwargs):
        from .elements.html_elements import Param
        return Param(self, dict(self._extract_selector(*args, **kwargs), tag_name='param'))

    def params(self, *args, **kwargs):
        from .elements.html_elements import ParamCollection
        return ParamCollection(self, dict(self._extract_selector(*args, **kwargs),
                                          tag_name='param'))

    def pre(self, *args, **kwargs):
        from .elements.html_elements import Pre
        return Pre(self, dict(self._extract_selector(*args, **kwargs), tag_name='pre'))

    def pres(self, *args, **kwargs):
        from .elements.html_elements import PreCollection
        return PreCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='pre'))

    def progress(self, *args, **kwargs):
        from .elements.html_elements import Progress
        return Progress(self, dict(self._extract_selector(*args, **kwargs), tag_name='progress'))

    def progresses(self, *args, **kwargs):
        from .elements.html_elements import ProgressCollection
        return ProgressCollection(self, dict(self._extract_selector(*args, **kwargs),
                                             tag_name='progress'))

    def q(self, *args, **kwargs):
        from .elements.html_elements import Quote
        return Quote(self, dict(self._extract_selector(*args, **kwargs), tag_name='q'))

    def qs(self, *args, **kwargs):
        from .elements.html_elements import QuoteCollection
        return QuoteCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='q'))

    def radio(self, *args, **kwargs):
        from .elements.radio import Radio
        return Radio(self, dict(self._extract_selector(*args, **kwargs), tag_name='input',
                                type='radio'))

    def radios(self, *args, **kwargs):
        from .elements.radio import RadioCollection
        return RadioCollection(self, dict(self._extract_selector(*args, **kwargs),
                                          tag_name='input', type='radio'))

    def radio_set(self, *args, **kwargs):
        from .elements.radio_set import RadioSet
        return RadioSet(self, dict(self._extract_selector(*args, **kwargs),
                                   tag_name='input', type='radio'))

    def rp(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='rp'))

    def rps(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='rp'))

    def rt(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='rt'))

    def rts(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='rt'))

    def ruby(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='ruby'))

    def rubies(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='ruby'))

    def s(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='s'))

    def ss(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='s'))

    def samp(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='samp'))

    def samps(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='samp'))

    def script(self, *args, **kwargs):
        from .elements.html_elements import Script
        return Script(self, dict(self._extract_selector(*args, **kwargs), tag_name='script'))

    def scripts(self, *args, **kwargs):
        from .elements.html_elements import ScriptCollection
        return ScriptCollection(self, dict(self._extract_selector(*args, **kwargs),
                                           tag_name='script'))

    def section(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='section'))

    def sections(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='section'))

    def select(self, *args, **kwargs):
        from .elements.select import Select
        return Select(self, dict(self._extract_selector(*args, **kwargs), tag_name='select'))

    select_list = select

    def selects(self, *args, **kwargs):
        from .elements.html_elements import SelectCollection
        return SelectCollection(self, dict(self._extract_selector(*args, **kwargs),
                                           tag_name='select'))

    select_lists = selects

    def small(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='small'))

    def smalls(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='small'))

    def source(self, *args, **kwargs):
        from .elements.html_elements import Source
        return Source(self, dict(self._extract_selector(*args, **kwargs), tag_name='source'))

    def sources(self, *args, **kwargs):
        from .elements.html_elements import SourceCollection
        return SourceCollection(self, dict(self._extract_selector(*args, **kwargs),
                                           tag_name='source'))

    def span(self, *args, **kwargs):
        from .elements.html_elements import Span
        return Span(self, dict(self._extract_selector(*args, **kwargs), tag_name='span'))

    def spans(self, *args, **kwargs):
        from .elements.html_elements import SpanCollection
        return SpanCollection(self, dict(self._extract_selector(*args, **kwargs),
                                         tag_name='span'))

    def strong(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='strong'))

    def strongs(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='strong'))

    def style(self, *args, **kwargs):
        from .elements.html_elements import Style
        return Style(self, dict(self._extract_selector(*args, **kwargs), tag_name='style'))

    def styles(self, *args, **kwargs):
        from .elements.html_elements import StyleCollection
        return StyleCollection(self, dict(self._extract_selector(*args, **kwargs),
                                          tag_name='style'))

    def sub(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='sub'))

    def subs(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='sub'))

    def summary(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='summary'))

    def summaries(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='summary'))

    def sup(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='sup'))

    def sups(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='sup'))

    def table(self, *args, **kwargs):
        from .elements.table import Table
        return Table(self, dict(self._extract_selector(*args, **kwargs), tag_name='table'))

    def tables(self, *args, **kwargs):
        from .elements.html_elements import TableCollection
        return TableCollection(self, dict(self._extract_selector(*args, **kwargs),
                                          tag_name='table'))

    def tbody(self, *args, **kwargs):
        from .elements.table_section import TableSection
        return TableSection(self, dict(self._extract_selector(*args, **kwargs), tag_name='tbody'))

    def tbodys(self, *args, **kwargs):
        from .elements.html_elements import TableSectionCollection
        return TableSectionCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                 tag_name='tbody'))

    def td(self, *args, **kwargs):
        from .elements.table_data_cell import TableDataCell
        return TableDataCell(self, dict(self._extract_selector(*args, **kwargs), tag_name='td'))

    def tds(self, *args, **kwargs):
        from .elements.html_elements import TableDataCellCollection
        return TableDataCellCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                  tag_name='td'))

    def template(self, *args, **kwargs):
        from .elements.html_elements import Template
        return Template(self, dict(self._extract_selector(*args, **kwargs), tag_name='template'))

    def templates(self, *args, **kwargs):
        from .elements.html_elements import TemplateCollection
        return TemplateCollection(self, dict(self._extract_selector(*args, **kwargs),
                                             tag_name='template'))

    def textarea(self, *args, **kwargs):
        from .elements.text_area import TextArea
        return TextArea(self, dict(self._extract_selector(*args, **kwargs), tag_name='textarea'))

    def textareas(self, *args, **kwargs):
        from .elements.html_elements import TextAreaCollection
        return TextAreaCollection(self, dict(self._extract_selector(*args, **kwargs),
                                             tag_name='textarea'))

    def text_field(self, *args, **kwargs):
        from .elements.text_field import TextField
        return TextField(self, dict(self._extract_selector(*args, **kwargs), tag_name='input'))

    def text_fields(self, *args, **kwargs):
        from .elements.text_field import TextFieldCollection
        return TextFieldCollection(self, dict(self._extract_selector(*args, **kwargs),
                                              tag_name='input'))

    def tfoot(self, *args, **kwargs):
        from .elements.table_section import TableSection
        return TableSection(self, dict(self._extract_selector(*args, **kwargs), tag_name='tfoot'))

    def tfoots(self, *args, **kwargs):
        from .elements.html_elements import TableSectionCollection
        return TableSectionCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                 tag_name='tfoot'))

    def th(self, *args, **kwargs):
        from .elements.html_elements import TableHeaderCell
        return TableHeaderCell(self, dict(self._extract_selector(*args, **kwargs), tag_name='th'))

    def ths(self, *args, **kwargs):
        from .elements.html_elements import TableHeaderCellCollection
        return TableHeaderCellCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                    tag_name='th'))

    def thead(self, *args, **kwargs):
        from .elements.table_section import TableSection
        return TableSection(self, dict(self._extract_selector(*args, **kwargs), tag_name='thead'))

    def theads(self, *args, **kwargs):
        from .elements.html_elements import TableSectionCollection
        return TableSectionCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                 tag_name='thead'))

    def time(self, *args, **kwargs):
        from .elements.html_elements import Time
        return Time(self, dict(self._extract_selector(*args, **kwargs), tag_name='time'))

    def times(self, *args, **kwargs):
        from .elements.html_elements import TimeCollection
        return TimeCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='time'))

    def title(self, *args, **kwargs):
        from .elements.html_elements import Title
        return Title(self, dict(self._extract_selector(*args, **kwargs), tag_name='title'))

    def titles(self, *args, **kwargs):
        from .elements.html_elements import TitleCollection
        return TitleCollection(self, dict(self._extract_selector(*args, **kwargs),
                                          tag_name='title'))

    def tr(self, *args, **kwargs):
        from .elements.table_row import TableRow
        return TableRow(self, dict(self._extract_selector(*args, **kwargs), tag_name='tr'))

    def trs(self, *args, **kwargs):
        from .elements.html_elements import TableRowCollection
        return TableRowCollection(self, dict(self._extract_selector(*args, **kwargs),
                                             tag_name='tr'))

    def track(self, *args, **kwargs):
        from .elements.html_elements import Track
        return Track(self, dict(self._extract_selector(*args, **kwargs), tag_name='track'))

    def tracks(self, *args, **kwargs):
        from .elements.html_elements import TrackCollection
        return TrackCollection(self, dict(self._extract_selector(*args, **kwargs),
                                          tag_name='track'))

    def u(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='u'))

    def us(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='u'))

    def ul(self, *args, **kwargs):
        from .elements.list import UList
        return UList(self, dict(self._extract_selector(*args, **kwargs), tag_name='ul'))

    def uls(self, *args, **kwargs):
        from .elements.html_elements import UListCollection
        return UListCollection(self, dict(self._extract_selector(*args, **kwargs), tag_name='ul'))

    def var(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='var'))

    def vars(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='var'))

    def video(self, *args, **kwargs):
        from .elements.html_elements import Video
        return Video(self, dict(self._extract_selector(*args, **kwargs), tag_name='video'))

    def videos(self, *args, **kwargs):
        from .elements.html_elements import VideoCollection
        return VideoCollection(self, dict(self._extract_selector(*args, **kwargs),
                                          tag_name='video'))

    def wbr(self, *args, **kwargs):
        from .elements.html_elements import HTMLElement
        return HTMLElement(self, dict(self._extract_selector(*args, **kwargs), tag_name='wbr'))

    def wbrs(self, *args, **kwargs):
        from .elements.html_elements import HTMLElementCollection
        return HTMLElementCollection(self, dict(self._extract_selector(*args, **kwargs),
                                                tag_name='wbr'))



from importlib import import_module

import six
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support.event_firing_webdriver import EventFiringWebDriver

import nerodia
from . import locators
from .after_hooks import AfterHooks
from .alert import Alert
from .capabilities import Capabilities
from .container import Container
from .cookies import Cookies
from .exception import Error, NoMatchingWindowFoundException
from .has_window import HasWindow
from .wait.wait import Waitable

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse


class Browser(Container, HasWindow, Waitable):
    def __init__(self, browser='chrome', *args, **kwargs):
        """
        Creates a nerodia.browser.Browser instance
        :param browser: firefox, ie, chrome, remote or Selenium WebDriver instance
        :type browser: selenium.webdriver.remote.webdriver.WebDriver or str
        :param args: args passed to the underlying driver
        :param kwargs: kwargs passed to the underlying driver
        """
        if isinstance(browser, six.string_types[0]):
            module = import_module('selenium.webdriver.{}.webdriver'.format(browser.lower()))
            caps = Capabilities(browser, **kwargs)
            self.driver = module.WebDriver(**caps.kwargs)
        elif isinstance(browser, WebDriver):
            self.driver = browser
        else:
            raise TypeError('A browser name or WebDriver instance must be supplied, '
                            'got {}'.format(type(browser)))

        if 'listener' in kwargs:
            self.driver = EventFiringWebDriver(self.driver, kwargs.get('listener'))

        self.after_hooks = AfterHooks(self)
        self.current_frame = None
        self.closed = False
        self.default_context = True
        self._original_window = None
        self._locator_namespace = locators

    @property
    def locator_namespace(self):
        """
        Whether the locators should be used from a different namespace. Defaults to nerodia.locators
        :return:
        """
        return self._locator_namespace

    @locator_namespace.setter
    def locator_namespace(self, namespace):
        self._locator_namespace = namespace

    @property
    def wd(self):
        return self.driver

    @staticmethod
    def start(url, browser='chrome', *args, **kwargs):
        """
        Creates a Browser instance
        :param url: url to navigate to after starting browser
        :type url: str
        :param browser: firefox, ie, chrome, remote or Selenium WebDriver instance
        :type browser: selenium.webdriver.remote.webdriver.WebDriver or str
        :param args: args passed to the underlying driver
        :param kwargs: kwargs passed to the underlying driver
        """
        b = Browser(browser, *args, **kwargs)
        b.goto(url)
        return b

    def __repr__(self):
        try:
            if self.alert.exists:
                return '#<{}:0x{:x} alert=True>'.format(self.__class__.__name__,
                                                        self.__hash__() * 2)
            else:
                return '#<{}:0x{:x} url={!r} title={!r}>'.format(self.__class__.__name__,
                                                                 self.__hash__() * 2, self.url,
                                                                 self.title)
        except:  # noqa
            return '#<{}:0x{:x} closed={}>'.format(self.__class__.__name__, self.__hash__() * 2,
                                                   self.closed)

    selector_string = __repr__

    def goto(self, uri):
        """
        Goes to the given URL
        :param uri: the URL
        :type uri: str
        :return: the url you end up at
        :rtype: str
        """
        scheme = urlparse(uri).scheme
        if scheme == '' or '.' in scheme:
            uri = 'http://{}'.format(uri)

        self.driver.get(uri)
        self.after_hooks.run()
        return uri

    def back(self):
        """ Navigates back in history """
        self.driver.back()
        self.after_hooks.run()

    def forward(self):
        """ Navigates forward in history """
        self.driver.forward()
        self.after_hooks.run()

    @property
    def url(self):
        """
        Returns the URL of the current page
        :rtype: str
        """
        self.assert_exists()
        return self.driver.current_url

    @property
    def title(self):
        """
        Returns the title of the current page
        :rtype: str
        """
        return self.driver.title

    def close(self):
        """ Closes the browser """
        if not self.closed:
            self.driver.quit()
            self.closed = True

    quit = close

    @property
    def cookies(self):
        """
        Handles cookies
        :rtype: nerodia.cookies.Cookies
        """
        return Cookies(self.driver)

    @property
    def name(self):
        """
        Returns the browser name
        :rtype: str
        """
        return self.driver.name

    @property
    def text(self):
        """
        Returns the text of the page body
        :return:
        """
        return self.body().text

    @property
    def html(self):
        """
        Returns HTML code of the current page
        :rtype: str
        """
        return self.driver.page_source

    @property
    def alert(self):
        """
        Handles Javascript alerts, confirms and prompts
        :rtype: nerodia.alert.Alert
        """
        return Alert(self)

    def refresh(self):
        """ Refreshes the current page """
        self.driver.refresh()
        self.after_hooks.run()

    def wait(self, timeout=5):
        """
        Waits until the readyState of document is complete, raises a TimeoutException if timeout is
        exceeded
        :param timeout: time to wait
        :type timeout: int
        """
        return self.wait_until(lambda b: b.ready_state == "complete", timeout=timeout,
                               message="waiting for document.readyState == 'complete'")

    @property
    def ready_state(self):
        """
        Returns the readyState of the document
        :rtype: str
        """
        return self.execute_script('return document.readyState;')

    @property
    def status(self):
        """
        Returns the text of the status bar
        :return:
        """
        return self.execute_script('return window.status;')

    def execute_script(self, script, *args):
        """
        Executes JavaScript snippet
        :param script: Javascript Snippet to execute
        :type script: str
        :param args: Arguments will be available in the given script in the 'arguments' pseudo-array
        :return: result of script
        """
        from .elements.element import Element
        args = [e.wd if isinstance(e, Element) else e for e in args]
        returned = self.driver.execute_script(script, *args)

        return self._wrap_elements_in(self, returned)

    def send_keys(self, *args):
        """
        Sends sequence of keystrokes to currently active element
        :param args: keystrokes
        """
        self.driver.switch_to.active_element.send_keys(*args)

    @property
    def screenshot(self):
        """
        Handles screenshots of current pages
        :rtype: nerodia.screenshot.Screenshot
        """
        from .screenshot import Screenshot
        return Screenshot(self)

    @property
    def exist(self):
        """
        True if browser is not closed and False otherwise
        :rtype: bool
        """
        return not self.closed and self.window().present

    exists = exist

    @property
    def browser(self):
        return self

    def assert_exists(self):
        self._ensure_context()
        if self.window().present:
            return
        raise NoMatchingWindowFoundException('browser window was closed')

    # private

    def _ensure_context(self):
        if self.closed:
            raise Error('browser was closed')
        if not self.default_context:
            self.driver.switch_to.default_content()
        self.default_context = True
        return self.default_context

    @staticmethod
    def _wrap_elements_in(scope, obj):
        if isinstance(obj, WebElement):
            return Browser._wrap_element(scope, obj)
        elif isinstance(obj, list):
            return [Browser._wrap_elements_in(scope, e) for e in obj]
        elif isinstance(obj, dict):
            for k, v in obj.items():
                obj[k] = Browser._wrap_elements_in(scope, v)
            return obj
        else:
            return obj

    @staticmethod
    def _wrap_element(scope, element):
        from .elements.html_elements import HTMLElement
        klass = nerodia.element_class_for(element.tag_name.lower()) or HTMLElement
        return klass(scope, {'element': element})



import re

from ..element.selector_builder import SelectorBuilder as ElementSelectorBuilder


class SelectorBuilder(ElementSelectorBuilder):
    # private

    def _normalize_selector(self, how, what):
        # We need to iterate through located elements and fetch
        # attribute value using Selenium because XPath doesn't understand
        # difference between IDL vs content attribute.
        # Current Element design doesn't allow to do that in any
        # obvious way except to use regular expression.

        if how == 'value' and isinstance(what, str):
            return [how, re.compile(r'^{}$'.format(re.escape(what)))]
        else:
            return super(SelectorBuilder, self)._normalize_selector(how, what)



import six

from .html_elements import InputCollection
from .i_frame import IFrame
from .input import Input
from ..browser import Browser
from ..meta_elements import MetaHTMLElement
from ..user_editable import UserEditable


@six.add_metaclass(MetaHTMLElement)
class TextField(UserEditable, Input):
    NON_TEXT_TYPES = ['file', 'radio', 'checkbox', 'submit', 'reset', 'image', 'button', 'hidden',
                      'range', 'color', 'date', 'datetime-local']

    @property
    def selector_string(self):
        selector = self.selector.copy()
        selector['type'] = '(any text type)'
        selector['tag_name'] = 'input'

        if isinstance(self.query_scope, Browser) or isinstance(self.query_scope, IFrame):
            return super(TextField, self).selector_string
        else:
            return '{} --> {}'.format(self.query_scope.selector_string, selector)


@six.add_metaclass(MetaHTMLElement)
class TextFieldCollection(InputCollection):
    # private
    @property
    def _element_class(self):
        return TextField



import logging


class Logger(object):
    def __init__(self):
        self._logger = logging.getLogger(__name__.split('.')[0])
        self.level = logging.WARNING
        self._filename = None
        self._ignored = []

    @property
    def level(self):
        return self._logger.level

    @level.setter
    def level(self, severity):
        if type(severity) == str:
            severity = severity.upper()
        self._logger.setLevel(severity)

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, path):
        self._filename = path
        for hdlr in self._logger.handlers[:]:  # remove all old handlers
            hdlr.close()
            self._logger.removeHandler(hdlr)

        if path is not None:
            fileh = logging.FileHandler(path, 'a')
            self._logger.addHandler(fileh)

    def ignore(self, *ids):
        for _id in ids:
            self._ignored.append(_id)

    def warning(self, msg, ids=None, *args, **kwargs):
        ids = ids or []
        if ids:
            message = '[{}]'.format(', '.join(ids))
        else:
            message = ''
        message += msg
        if len(set(self._ignored).intersection(ids)) == 0:
            self._logger.warning(message, *args, **kwargs)

    warn = warning

    def deprecate(self, old, new, ids=None):
        ids = ids or []
        if 'deprecations' in self._ignored or set(self._ignored).intersection(ids):
            return
        if ids:
            message = '[{}]'.format(', '.join(ids))
        else:
            message = ''
        self.warning('[DEPRECATION] {}{} is deprecated. Use {} instead.'.format(message, old, new))

    def __getattr__(self, item):
        return getattr(self._logger, item)



import six

from .html_elements import HTMLElement
from ..meta_elements import MetaHTMLElement


@six.add_metaclass(MetaHTMLElement)
class Option(HTMLElement):
    # alias
    def select(self):
        self.click()

    toggle = select

    def clear(self):
        """ Un-selects the option """
        if self.is_selected:
            self.click()

    @property
    def is_selected(self):
        """
        Returns True if the option is selected
        :rtype: bool
        """
        return self._element_call(lambda: self.el.is_selected())

    @property
    def text(self):
        """
        Returns the text of the option
        Note that the text is either one of the following respectively:
            * label attribute
            * text attribute
            * inner element text
        :rtype: str
        """
        for attr in ['label', 'text']:
            val = self.attribute_value(attr)
            if val:
                return val

        return super(Option, self).text



from selenium.webdriver import DesiredCapabilities, ChromeOptions, FirefoxOptions

import nerodia


class Capabilities(object):
    DEFAULT_URL = 'http://127.0.0.1:{}/wd/hub'

    def __init__(self, browser, **options):
        self.options = options.copy()
        nerodia.logger.info('Creating Browser instance of {} with user provided options: '
                            '{}'.format(browser, options))

        if browser == 'remote' and 'browser' in self.options:
            self.browser = self.options.pop('browser')
        elif browser == 'remote' and 'desired_capabilities' in self.options:
            self.browser = self.options['desired_capabilities'].get('browserName')
        else:
            self.browser = browser

        if browser == 'remote' or options.get('url'):
            self.selenium_browser = 'remote'
        else:
            self.selenium_browser = browser

        self.selenium_opts = {}

    @property
    def kwargs(self):
        self._process_arguments()
        return self.selenium_opts

    # private

    def _process_arguments(self):
        url = None
        port = self.options.pop('port', None)
        if port:
            url = self.DEFAULT_URL.format(port)

        url = self.options.pop('url', url)
        if url:
            self.selenium_opts['command_executor'] = url

        self._process_browser_options()
        self._process_capabilities()
        nerodia.logger.info('Creating Browser instance with Nerodia processed options: '
                            '{}'.format(self.selenium_opts))

    def _process_browser_options(self):
        browser_options = self.options.pop('options', {})

        if self.selenium_browser == 'chrome':
            self._process_chrome_options(browser_options)
        elif self.selenium_browser == 'firefox':
            self._process_firefox_options(browser_options)
        elif self.selenium_browser == 'safari':
            self._process_safari_options(browser_options)
        elif self.selenium_browser == 'ie':
            self._process_ie_options(browser_options)
        elif self.selenium_browser == 'remote':
            self._process_remote_options(browser_options)

    def _process_chrome_options(self, opts):
        if isinstance(opts, ChromeOptions):
            options = opts
        else:
            options = ChromeOptions()
            if 'args' in opts:
                for arg in opts.pop('args'):
                    options.add_argument(arg)
            if 'headless' in opts or 'headless' in self.options:
                options.set_headless()
            if 'binary' in opts or 'binary_location' in opts:
                options.binary_location = opts.pop('binary') or opts.pop('binary_location')
            if 'debugger_address' in opts:
                options.debugger_address = opts.pop('debugger_address')
            if 'extensions' in opts:
                for path in opts.pop('extensions'):
                    options.add_extension(path)
            if 'encoded_extensions' in opts:
                for string in opts.pop('encoded_extensions'):
                    options.add_encoded_extension(string)
            if 'experimental_options' in opts:
                for name, value in opts.pop('experimental_options').items():
                    options.add_experimental_option(name, value)
        self.selenium_opts['options'] = options

    def _process_firefox_options(self, opts):
        if isinstance(opts, FirefoxOptions):
            options = opts
        else:
            options = FirefoxOptions()
            if 'args' in opts:
                for arg in opts.pop('args'):
                    options.add_argument(arg)
            if 'headless' in opts or 'headless' in self.options:
                options.set_headless()
            if 'binary' in opts or 'binary_location' in opts:
                options.binary = opts.pop('binary') or opts.pop('binary_location')
            if 'prefs' in opts:
                for name, value in opts.pop('prefs').items():
                    options.set_preference(name, value)
            if 'proxy' in opts:
                options.proxy = opts.pop('proxy')
            if 'profile' in opts:
                options.profile = opts.pop('profile')
            if 'log_level' in opts:
                options.log.level = opts.pop('log_level')
        self.selenium_opts['options'] = options

    def _process_safari_options(self, opts):
        if 'technology_preview' in opts:
            self.selenium_opts['executable_path'] = \
                '/Applications/Safari Technology Preview.app/Contents/MacOS/safaridriver'

    def _process_ie_options(self, opts):
        from selenium.webdriver.ie.options import Options
        if isinstance(opts, Options):
            options = opts
        else:
            options = Options()
            if 'args' in opts:
                for arg in opts.pop('args'):
                    options.add_argument(arg)
        self.selenium_opts['options'] = options

    def _process_remote_options(self, opts):
        if self.browser == 'chrome' and self.options.pop('headless', None):
            if 'desired_capabilities' in opts:
                caps = opts.pop('desired_capabilities')
            else:
                caps = DesiredCapabilities.CHROME.copy()
            self._process_chrome_options(opts)
            self.selenium_opts['desired_capabilities'] = caps

        if self.browser == 'firefox' and self.options.pop('headless', None):
            if 'desired_capabilities' in opts:
                caps = opts.pop('desired_capabilities')
            else:
                caps = DesiredCapabilities.FIREFOX.copy()
            self._process_firefox_options(opts)
            self.selenium_opts['desired_capabilities'] = caps

        if self.browser == 'safari' and self.options.pop('technology_preview', None):
            if 'desired_capabilities' in opts:
                caps = opts.pop('desired_capabilities')
            else:
                caps = DesiredCapabilities.SAFARI.copy()
            caps['safari.options'] = {'technologyPreview': opts.pop('technology_preview')}
            self.selenium_opts['desired_capabilities'] = caps

    def _process_capabilities(self):
        caps = self.options.pop('desired_capabilities', None)

        if caps:
            nerodia.logger.warn('You can now pass values directly into nerodia.browser.Browser '
                                'without needing to use desired_capabilities',
                                ids=['use_capabilities'])
            self.selenium_opts.update(self.options)
        else:
            from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
            browser = 'internetexplorer' if self.browser == 'ie' else self.browser
            caps = getattr(DesiredCapabilities, browser.upper())

        if self.browser in ['firefox', 'ie', 'edge']:
            self.selenium_opts['capabilities'] = caps
        else:
            self.selenium_opts['desired_capabilities'] = caps



try:
    from re import Pattern
except ImportError:
    from re import _pattern_type as Pattern


class Adjacent(object):
    def parent(self, **kwargs):
        """
        Returns parent element of current element
        :rtype: nerodia.elements.elements.Element

        :Example:

        browser.text_field(name='new_user_field_name').parent() == browser.fieldset    #=> True
        """
        kwargs['index'] = kwargs.get('index', 0)
        return self._xpath_adjacent(**dict(kwargs, adjacent='ancestor', plural=False))

    def preceding_sibling(self, **kwargs):
        """
        Returns preceding sibling element of current element
        :rtype: nerodia.elements.elements.Element

        :Example:

        browser.text_field(name='new_user_first_name').preceding_sibling(index=1) == browser.legend
        #=> True
        """
        return self._xpath_adjacent(**dict(kwargs, adjacent='preceding', plural=False))

    previous_sibling = preceding_sibling

    def preceding_siblings(self, **kwargs):
        """
        Returns collection of preceding sibling elements of current element
        :rtype: nerodia.element_collection.ElementCollection

        :Example:

        len(browser.text_field(name='new_user_first_name').preceding_siblings)    #=> 3
        """
        if 'index' in kwargs:
            raise ValueError('#previous_siblings can not take an index value')
        return self._xpath_adjacent(**dict(kwargs, adjacent='preceding', plural=True))

    previous_siblings = preceding_siblings

    def following_sibling(self, **kwargs):
        """
        Returns following sibling element of current element
        :rtype: nerodia.elements.elements.Element

        :Example:

        browser.text_field(name='new_user_first_name').following_sibling(index=2) == \
            browser.text_field(id='new_user_last_name')    #=> True
        """
        return self._xpath_adjacent(**dict(kwargs, adjacent='following', plural=False))

    next_sibling = following_sibling

    def following_siblings(self, **kwargs):
        """
        Returns collection of following sibling elements of current element
        :rtype: nerodia.element_collection.ElementCollection

        :Example:

        len(browser.text_field(name='new_user_first_name').following_siblings)    #=> 52
        """
        if 'index' in kwargs:
            raise ValueError('#following_siblings can not take an index value')
        return self._xpath_adjacent(**dict(kwargs, adjacent='following', plural=True))

    next_siblings = following_siblings

    def siblings(self, **kwargs):
        """
        Returns collection of sibling elements of current element, including the current element
        :rtype: nerodia.element_collection.ElementCollection

        :Example:

        len(browser.text_field(name='new_user_first_name').siblings)    #=> 56
        """
        return self.parent().children(**kwargs)

    def child(self, **kwargs):
        """
        Returns element of direct child of current element
        :rtype: nerodia.elements.elements.Element

        :Example:

        browser.form(id='new_user').child == browser.fieldset    #=> True
        """
        return self._xpath_adjacent(**dict(kwargs, adjacent='child', plural=False))

    def children(self, **kwargs):
        """
        Returns collection of elements of direct children of current element
        :rtype: nerodia.element_collection.ElementCollection

        :Example:

        browser.select_list(id='new_user_languages').children == \
            list(browser.select_list(id='new_user_languages').options)    #=> True
        """
        if 'index' in kwargs:
            raise ValueError('#children can not take an index value')
        return self._xpath_adjacent(**dict(kwargs, adjacent='child', plural=True))

    # private

    def _xpath_adjacent(self, **kwargs):
        from .elements.html_elements import HTMLElement, HTMLElementCollection
        import nerodia

        plural = kwargs.pop('plural', None)
        index = kwargs.pop('index', None)
        tag_name = kwargs.get('tag_name')

        if not (plural or any(isinstance(val, Pattern) for val in kwargs.values())):
            kwargs['index'] = index or 0

        if not plural and tag_name:
            klass = nerodia.tag_to_class.get(tag_name)
        elif not plural:
            klass = HTMLElement
        elif tag_name:
            klass = nerodia.tag_to_class.get('{}_collection'.format(tag_name))
        else:
            klass = HTMLElementCollection

        return klass(self, kwargs)



from ..element.validator import Validator as ElementValidator


class Validator(ElementValidator):
    @staticmethod
    def validate(element, selector):
        if element.tag_name.lower() == 'input':
            return element



from selenium.webdriver import Remote

import nerodia


class Screenshot(object):

    def __init__(self, browser):
        if isinstance(browser, Remote):
            nerodia.logger.deprecate('Initializing `Screenshot` with a `selenium.webdriver` '
                                     'instance', 'a `nerodia.browser` instance',
                                     ids=['screenshot_driver'])
            self.driver = browser
        else:
            self.browser = browser
            self.driver = browser.wd

    def save(self, path):
        """
        Saves screenshot to given path
        :param path: file path

        :Example:

        browser.screenshot.save('screenshot.png')
        """
        self.driver.save_screenshot(path)

    def png(self):
        """
        Represents screenshot as PNG image string
        :rtype: str

        :Example:

        browser.screenshot.png
        #=> '\x95\xC7\x8C@1\xC07\x1C(Edb\x15\xB2\vL'
        """
        return self.driver.get_screenshot_as_png()

    def base64(self):
        """
        Represents screenshot as Base64 encoded string
        :rtype: str

        :Example:

        browser.screenshot.base64
        #=> '7HWJ43tZDscPleeUuPW6HhN3x+z7vU/lufmH0qNTtTum94IBWMT46evImci1vnFGT'
        """
        return self.driver.get_screenshot_as_base64()



from selenium.common.exceptions import NoAlertPresentException

import nerodia
from .exception import UnknownObjectException
from .wait.wait import Waitable, TimeoutError


class Alert(Waitable):
    def __init__(self, browser):
        self.browser = browser
        self.alert = None

    @property
    def text(self):
        """
        Returns the text of the alert
        :rtype: str

        :Example:

        browser.alert.text    #=> 'ok'
        """
        self.wait_for_exists()
        return self.alert.text

    def ok(self):
        """
        Closes alert or accepts prompts/confirms

        :Example:

        browser.alert.ok
        browser.alert.exists    #=> False
        """
        self.wait_for_exists()
        self.alert.accept()
        self.browser.after_hooks.run()

    def close(self):
        """
        Closes alert or cancels prmopts/confirms

        :Example:

        browser.alert.close()
        browser.alert.exists    #=> False
        """
        self.wait_for_exists()
        self.alert.dismiss()
        self.browser.after_hooks.run()

    def set(self, value):
        """
        Enters text to prompt
        :param value: keys to send

        :Example:

        browser.alert.set('Text for prompt')
        browser.alert.ok()
        """
        self.wait_for_exists()
        self.alert.send_keys(value)

    @property
    def exists(self):
        """
        Returns True if alert, confirm, or prompt is present and False otherwise
        :rtype: bool

        :Example:

        browser.alert.exists    #=> True
        """
        try:
            self.assert_exists()
            return True
        except UnknownObjectException:
            return False

    present = exists

    @property
    def selector_string(self):
        return 'alert'

    def assert_exists(self):
        try:
            self.alert = self.browser.driver.switch_to.alert
        except NoAlertPresentException:
            raise UnknownObjectException('unable to locate alert')

    def wait_for_exists(self):
        if not nerodia.relaxed_locate:
            return self.assert_exists()

        try:
            return self.wait_until(lambda a: a.exists, message='waiting for alert')
        except TimeoutError:
            if nerodia.default_timeout != 0:
                msg = 'This code has slept for the duration of the default timeout waiting for ' \
                      'an Alert to exist. If the test is still passing, consider using ' \
                      'Alert#exists? instead of rescuing UnknownObjectException'
                nerodia.logger.warning(msg, ids=['wait_for_alert'])
            raise UnknownObjectException('unable to locate alert')



from datetime import date, datetime
from inspect import stack

import six
from dateutil import parser

from nerodia.elements.html_elements import InputCollection
from .input import Input
from ..meta_elements import MetaHTMLElement


@six.add_metaclass(MetaHTMLElement)
class DateField(Input):

    def js_set(self, value):
        """
        Set date field to the given date

        :param value: value to set to
        :type value: datetime or date or str

        :Example:

        browser.date_field(id='start_date').set('2018/01/31')

        """
        if isinstance(value, six.string_types):
            value = parser.parse(value, fuzzy=True)

        if not isinstance(value, (date, datetime)):
            raise TypeError('DateField#{} only accepts instances of date or '
                            'datetime'.format(stack()[0][3]))

        date_string = value.strftime('%Y-%m-%d')
        self._element_call(lambda: self._execute_js('setValue', self.el, date_string),
                           precondition=self.wait_for_writable)

    set = js_set

    @property
    def value(self):
        return self.attribute_value('value')

    @value.setter  # alias
    def value(self, *args):
        self.set(*args)


class DateFieldCollection(InputCollection):

    # private

    @property
    def _element_class(self):
        return DateField



from selenium.common.exceptions import NoSuchWindowException, WebDriverException

from .exception import NoMatchingWindowFoundException
from .wait.wait import Waitable, TimeoutError


class Window(Waitable):
    def __init__(self, browser, selector):
        self.browser = browser
        self.driver = browser.driver
        self.selector = selector
        self.original = self._current_window
        self.window_handle = None

        if not selector:
            self.window_handle = self.original
        elif 'handle' in selector:
            self.window_handle = selector.pop('handle')
        else:
            if not all(key in ['title', 'url', 'index'] for key in selector.keys()):
                raise ValueError('invalid window selector: {}'.format(selector))

    def __repr__(self):
        return '#<{}:0x{:x} located={}>'.format(self.__class__.__name__, hash(self) * 2,
                                                self.handle is not None)

    def __enter__(self):
        self.original = self._current_window
        return self.use()

    def __exit__(self, *args):
        self.unuse()

    @property
    def size(self):
        """
        Returns window size
        :rtype: dict

        :Example:

        browser.window().size    #=> {width: 1600, height: 1200}
        """
        with self:
            size = self.driver.get_window_size()
        return Dimension(**size)

    @property
    def position(self):
        """
        Returns window position
        :rtype: dict

        :Example:

        browser.window().position    #=> {x: 92, y: 76}
        """
        with self:
            pos = self.driver.get_window_position()
        return Point(**pos)

    def resize_to(self, width, height):
        """
        Resizes window to given width and height
        :param width: width to resize to
        :param height: height to resize to

        :Example:

        browser.window().resize_to(1600, 1200)
        """
        size = {'width': width, 'height': height}
        with self:
            self.driver.set_window_size(**size)
        return size

    def move_to(self, x, y):
        """
        Moves window to given x and y coordinates
        :param x: horizontal position
        :param y: vertical position

        :Example:

        browser.window().move_to(300, 200)
        """
        point = {'x': x, 'y': y}
        with self:
            self.driver.set_window_position(**point)
        return point

    def maximize(self):
        """
        Maximizes window

        :Example:

        browser.window().maximize()
        """
        with self:
            self.driver.maximize_window()

    @property
    def exists(self):
        """
        Returns True if the window exists
        :rtype: bool
        """
        try:
            self.assert_exists()
            return True
        except NoMatchingWindowFoundException:
            return False

    present = exists
    exist = exists

    def __eq__(self, other):
        """
        Returns True if two windows are equal
        :param other: the other window
        :rtype: bool

        :Example:

        browser.window(index=0) == browser.window(index=1)    #=> False
        """
        if not isinstance(other, Window):
            return False

        return self.handle == other.handle

    eql = __eq__

    def __hash__(self):
        return hash(self.handle) ^ hash(self.__class__)

    @property
    def is_current(self):
        """
        Returns True if the window is current

        :Example:

        browse.window.is_current

        :rtype: bool
        """
        return self._current_window == self.handle

    def close(self):
        """ Closes the window """
        if self == self.browser.original_window:
            self.browser._original_window = None
        if self._current_window == self.handle:
            self.driver.close()
        else:
            with self:
                self.driver.close()

    @property
    def title(self):
        """
        Returns the window title
        :rtype: str
        """
        with self:
            title = self.driver.title
        return title

    @property
    def url(self):
        """
        Returns the window URL
        :rtype: str
        """
        with self:
            url = self.driver.current_url
        return url

    def use(self):
        """
        Switches to given window

        :Example:

        browser.window(title='closeable window').use()
        :rtype: Window
        """
        if self.browser.original_window is None:
            self.browser._original_window = self
        self.wait_for_exists()
        self.driver.switch_to.window(self.handle)
        return self

    def unuse(self):
        """
        Returns to the original window

        :Example:

        window = browser.window(title='closeable window').use()
        window.unuse()
        :rtype: Window
        """
        current = self.driver.window_handles
        orig = self.original if self.original in current else current[0]
        self.driver.switch_to.window(orig)
        return self

    @property
    def handle(self):
        return self.window_handle or self._locate()

    # Referenced in EventuallyPresent
    def selector_string(self):
        return str(self.selector)

    def assert_exists(self):
        if self.handle not in self.driver.window_handles:
            raise NoMatchingWindowFoundException(str(self.selector))

    def wait_for_exists(self):
        import nerodia
        if not nerodia.relaxed_locate:
            self.assert_exists()
        try:
            self.wait_until(lambda w: w.exists)
        except TimeoutError:
            raise NoMatchingWindowFoundException(str(self.selector))

    # private

    def _locate(self):
        if not self.selector:
            self.window_handle = None
        elif 'index' in self.selector:
            try:
                self.window_handle = self.driver.window_handles[int(self.selector.get('index'))]
            except IndexError:
                self.window_handle = None
        else:
            self.window_handle = next((x for x in self.driver.window_handles if
                                       self._matches(x) is True), False)
        return self.window_handle

    @property
    def _current_window(self):
        try:
            return self.driver.current_window_handle
        except WebDriverException:
            return None

    def _matches(self, handle):
        from .locators.element.validator import Validator
        try:
            orig = self.driver.current_window_handle
        except NoSuchWindowException:
            orig = None
        try:
            self.driver.switch_to.window(handle)

            if 'title' in self.selector:
                title_value = self.selector.get('title')
                driver_title = self.driver.title
                matches_title = Validator.match_str_or_regex(title_value, driver_title)
            else:
                matches_title = True

            if 'url' in self.selector:
                url_value = self.selector.get('url')
                driver_url = self.driver.current_url
                matches_url = Validator.match_str_or_regex(url_value, driver_url)
            else:
                matches_url = True

            return matches_title and matches_url
        except (NoSuchWindowException, WebDriverException):
            return False
        finally:
            current = self.driver.window_handles
            orig = orig if orig in current else current[0]
            self.driver.switch_to.window(orig)


class Dimension(object):
    __slots__ = ['width', 'height']

    def __init__(self, width=None, height=None):
        self.width = width
        self.height = height

    def __eq__(self, other):
        return self.width == other.width and self.height == other.height


class Point(object):
    __slots__ = ['x', 'y']

    def __init__(self, x=None, y=None):
        self.x = x
        self.y = y

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y



from .window import Window


class HasWindow(object):
    def windows(self, *args, **kwargs):
        """
        Returns browser windows list
        :rtype: list[Window]

        :Example:

        browser.windows(title='closeable window')
        """
        all = [Window(self, {'handle': handle}) for handle in self.driver.window_handles]

        if not args and not kwargs:
            return all
        else:
            return self._filter_windows(self._extract_selector(*args, **kwargs), all)

    def window(self, *args, **kwargs):
        """
        Returns browser window
        :rtype: Window

        :Example:

        browser.window(title='closeable window')
        """
        return Window(self, self._extract_selector(*args, **kwargs))

    @property
    def original_window(self):
        """
        Returns original window if defined, current window if not
        :rtype: Window

        :Example:

        browser.window(title='closeable window').use()
        browser.original_window.use()
        """
        if not self._original_window:
            self._original_window = self.window()
        return self._original_window

    # private

    def _filter_windows(self, selector, windows):
        from .locators.element.validator import Validator
        if not all(key in ['title', 'url'] for key in selector.keys()):
            raise ValueError('invalid window selector: {}'.format(selector))
        return [w for w in windows if all(Validator.match_str_or_regex(v, getattr(w, k)) for k, v in selector.items())]



import six

from .logger import Logger

__version__ = '0.9.0'

ttc = None

tag_to_class = ttc or {}

#
# Whether or not Watip should wait for an element to be found or present before taking an action.
# Defaults to true.
#

relaxed_locate = True

#
# Default wait time for wait methods.
#

default_timeout = 30

#
# Custom logger
#

logger = Logger()

#
# For py2/py3 compatibility
#

_str_types = (six.binary_type, six.text_type)


from . import tag_map  # noqa


def element_class_for(tag_name):
    return tag_to_class.get(tag_name)



import logging

from ..element.selector_builder import SelectorBuilder as ElementSelectorBuilder, \
    XPath as ElementXPath
from ...elements.button import Button
from ...exception import Error
from ...xpath_support import XpathSupport

try:
    from re import Pattern
except ImportError:
    from re import _pattern_type as Pattern


class SelectorBuilder(ElementSelectorBuilder):
    def _build_wd_selector(self, selectors):
        if any(isinstance(val, Pattern) for val in selectors.values()):
            return None

        if not selectors.pop('tag_name', None):
            raise Error('internal error: no tag_name?!')

        button_attr_exp = self.xpath_builder.attribute_expression('button', selectors)

        xpath = './/button'
        if button_attr_exp:
            xpath += '[{}]'.format(button_attr_exp)

        if selectors.get('type') is not False:
            if selectors.get('type') in [None, True]:
                selectors['type'] = Button.VALID_TYPES
            input_attr_exp = self.xpath_builder.attribute_expression('input', selectors)

            xpath += ' | .//input'
            xpath += '[{}]'.format(input_attr_exp)

        logging.debug({'build_wd_selector': xpath})

        return ['xpath', xpath]


class XPath(ElementXPath):
    @staticmethod
    def lhs_for(building, key):
        if building == 'input' and key == 'text':
            return '@value'
        else:
            return super(XPath, XPath).lhs_for(building, key)

    def equal_pair(self, building, key, value):
        if building == 'button' and key == 'value':
            # :value should look for both node text and @value attribute
            text = XpathSupport.escape(value)
            return '(text()={0} or @value={0})'.format(text)
        else:
            return super(XPath, self).equal_pair(building, key, value)



from importlib import import_module
from itertools import islice

import nerodia
from .locators.class_helpers import ClassHelpers


class ElementCollection(ClassHelpers):
    def __init__(self, query_scope, selector):
        self.query_scope = query_scope
        self.selector = selector
        self.generator = ()
        self.els = []
        self.locator = None
        self.elements = None

    def __iter__(self):
        """
        Yields each element in collection

        :rtype: iter

        :Example:

        divs = browser.divs(class='kls')
        for div in divs:
             print(div.text)
        """
        from .elements.html_elements import HTMLElement
        from .elements.input import Input
        dic = {}
        for idx, e in enumerate(self._elements):
            element = self._element_class(self.query_scope, dict(self.selector, index=idx,
                                                                 element=e))
            if element.__class__ in [HTMLElement, Input]:
                tag_name = element.tag_name
                dic[tag_name] = dic.get(tag_name, 0)
                dic[tag_name] += 1
                kls = nerodia.element_class_for(tag_name)
                new_selector = dict(self.selector, element=e, tag_name=tag_name,
                                    index=dic[tag_name] - 1)
                yield kls(self.query_scope, new_selector)
            else:
                yield element

    def __len__(self):
        """
        Returns the number of elements in the collection
        :rtype: int
        """
        return len([_ for _ in self])

    def __getitem__(self, idx):
        """
        Get the element at the given index or slice

        Any call to an ElementCollection including an adjacent selector
        can not be lazy loaded because it must store correct type

        Slices can only be lazy loaded if the indices are positive

        :param idx: index of wanted element, 0-indexed
        :type idx: int
        :return: instance of Element subclass
        :rtype: nerodia.elements.element.Element
        """
        if isinstance(idx, slice):
            if idx.start and idx.start < 0 or idx.stop and idx.stop < 0:
                return list(self)[idx.start:idx.stop]
            else:
                return list(islice(self, idx.start, idx.stop, idx.step))
        elif 'adjacent' in self.selector:
            try:
                return list(islice(self, idx + 1))[idx]
            except IndexError:
                return self._element_class(self.query_scope, {'invalid_locator': True})
        else:
            return self._element_class(self.query_scope, dict(self.selector, index=idx))

    @property
    def is_empty(self):
        """
        Returns True if no elements are found

        :Example:

        browser.select_list(name='new_user_languages').options(class_name='not_here').is_empty

        :Example:

        browser.select_list(name='new_user_languages').options(id='danish').is_empty

        :return: True if no elements are found
        :rtype: bool
        """
        return len(self) == 0

    @property
    def to_list(self):
        """
        This collection as a list

        :rtype: list[nerodia.elements.element.Element]
        """
        nerodia.logger.deprecate('ElementCollection.to_list', 'list(self)')
        return list(self)

    def locate(self):
        """
        Locate all elements and return self

        :rtype: ElementCollection
        """
        list(self)
        return self

    @property
    def browser(self):
        """
        Returns the browser of the current query_scope

        :rtype: nerodia.browser.Browser
        """
        return self.query_scope.browser

    def __eq__(self, other):
        """
        Returns true if two element collections are equal.

        :param other: other collection
        :rtype: bool

        :Example:

        browser.select_list(name='new_user_languages').options == \
            browser.select_list(id='new_user_languages').options   #=> True

        browser.select_list(name=;new_user_role').options == \
            browser.select_list(id='new_user_languages').options   #=> false
        """
        return list(self) == list(other)

    eql = __eq__

    # private

    @property
    def _elements(self):
        if self.locator is None:
            self.locator = self._build_locator()
        if self.elements is None:
            self.elements = self.locator.locate_all()
        return self.elements

    @property
    def _element_class(self):
        from .elements.svg_elements import SVGElementCollection
        from .elements.html_elements import HTMLElementCollection
        from .module_mapping import map_module
        name = self.__class__.__name__.replace('Collection', '')
        element_module = map_module(name)
        try:
            module = import_module('nerodia.elements.{}'.format(element_module))
        except ImportError:
            if isinstance(self, HTMLElementCollection):
                module = import_module('nerodia.elements.html_elements')
            elif isinstance(self, SVGElementCollection):
                module = import_module('nerodia.elements.svg_elements')
            else:
                raise TypeError(
                    'element class for {} could not be determined'.format(name))
        return getattr(module, name)



import six

from .html_elements import HTMLElement
from ..meta_elements import MetaHTMLElement


@six.add_metaclass(MetaHTMLElement)
class Form(HTMLElement):
    def submit(self):
        self._element_call(lambda: self.el.submit(), self.wait_for_present)
        self.browser.after_hooks.run()



from nerodia.exception import UnknownObjectException
from .radio import Radio, RadioCollection


class RadioSet(object):
    def __init__(self, query_scope, selector):
        if not isinstance(selector, dict):
            raise ValueError('invalid argument: {}'.format(selector))

        self.source = Radio(query_scope, selector)
        self.frame = self.source.parent(tag_name='form')

    def __iter__(self):
        """
        Yields each Radio associated with this set

        :rtype: iter

        :Example:

        radio_set = browser.radio_set(id='some_radio_set')
        for radio in radio_set:
            print(radio.text)
        """
        for e in self.radios():
            yield e

    def __len__(self):
        """
        Returns the number of Radios in the RadioSet
        :rtype: int
        """
        return len(self.radios())

    def __getitem__(self, idx):
        """
        Get the Radio at the given index in the RadioSet

        :param idx: index of Radio element, 0-indexed
        :type idx: int
        :return: instance of Radio
        :rtype: nerodia.elements.radio.Radio
        """
        return self.radios()[idx]

    def radio(self, **kwargs):
        """
        Gets a Radio for the RadioSet
        :rtype: Radio
        """
        name = self.name
        if name and ('name' not in kwargs or kwargs.get('name') == name):
            return self.frame.radio(**dict(kwargs, name=name))
        elif not name:
            return self.source
        else:
            raise UnknownObjectException('{} does not match name of RadioSet: '
                                         '{}'.format(kwargs.get('name'), name))

    def radios(self, **kwargs):
        """
        Gets a RadioCollection for the RadioSet
        :rtype: RadioCollection
        """
        name = self.name
        if name and ('name' not in kwargs or kwargs.get('name') == name):
            return self._element_call(lambda: self.frame.radios(**dict(kwargs, name=name)),
                                      self.source.wait_for_present)
        elif not name:
            return RadioCollection(self.frame, {'element': self.source.wd})
        else:
            raise UnknownObjectException('{} does not match name of RadioSet: '
                                         '{}'.format(kwargs.get('name'), name))

    @property
    def enabled(self):
        """
        Returns True if any radio buttons in the set are enabled

        :rtype: bool
        """
        return any(e.enabled for e in self)

    @property
    def disabled(self):
        """
        Returns True if all radio buttons in the set are disabled

        :rtype: bool
        """
        return not self.enabled

    @property
    def name(self):
        """
        Returns the name attribute for the set

        :rtype: str
        """
        return self.source.name

    @property
    def type(self):
        """
        Returns the name attribute for the set

        :rtype: str
        """
        self.source.assert_exists()
        return 'radio'

    def includes(self, term):
        """
        Returns True if the radio set has one ore more radio buttons where label matches the given
        value

        :rtype: bool
        """
        return self.radio(label=term).exists

    def select(self, term):
        """
        Select the radio button whose value or label matches the given string.

        :param term: string or regex
        :returns: the value or text of the radio selected
        :rtype: str
        """
        found_by_value = self.radio(value=term)
        found_by_text = self.radio(label=term)

        if found_by_value.exists:
            if not found_by_value.is_selected:
                found_by_value.click()
            return found_by_value.value
        elif found_by_text.exists:
            if not found_by_text.is_selected:
                found_by_text.click()
            return found_by_text.text
        else:
            raise UnknownObjectException('Unable to locate radio matching {}'.format(term))

    def is_selected(self, term):
        """
        Returns True if the radio button found matching the term is selected

        :param term: string or regex
        :rtype: bool
        """
        found = self.frame.radio(label=term)
        if found.exists:
            return found.is_selected
        raise UnknownObjectException('Unable to locate radio matching {}'.format(term))

    @property
    def selected(self):
        """
        Returns the selected Radio element
        Returns None if no radio button is selected
        :rtype: Radio or None
        """
        return next((el for el in self if el.is_selected), None)

    @property
    def value(self):
        """
        Returns the value of the selected radio button in the set.
        Returns None if no radio button is selected
        :rtype: str or None
        """
        sel = self.selected
        return sel and sel.value

    @property
    def text(self):
        """
        Returns the text of the selected radio button in the set.
        Returns None if no radio button is selected
        :rtype: str or None
        """
        sel = self.selected
        return sel and sel.text

    def __eq__(self, other):
        """
        Returns True if two RadioSets are equal.

        :param other: other RadioSet
        :rtype: bool

        :Example:

        browser.radio_set(id='new_user_newsletter_yes') == browser.radio_set(id='new_user_newsletter_yes')   #=> True
        browser.radio_set(id='new_user_newsletter_yes') == browser.radio_set(id='new_user_newsletter_no')   #=> False
        """
        if not isinstance(other, RadioSet):
            return False
        return self.radios() == other.radios()

    def __getattr__(self, item):
        return getattr(self.source, item)

    # private

    def _element_call(self, method, exist_check=None):
        return self.source._element_call(method, exist_check)



#!/usr/bin/env python

from distutils.command.install import INSTALL_SCHEMES
from os.path import dirname, join, abspath

from setuptools import setup
from setuptools.command.install import install

for scheme in INSTALL_SCHEMES.values():
    scheme['data'] = scheme['purelib']

setup_args = {
    'cmdclass': {'install': install},
    'name': 'nerodia',
    'version': '0.9.0',
    'license': 'MIT',
    'description': 'Python port of WATIR https://github.com/watir/watir',
    'long_description': open(join(abspath(dirname(__file__)), 'README.rst')).read(),
    'url': 'https://github.com/watir/nerodia',
    'classifiers': ['Intended Audience :: Developers',
                    'Operating System :: POSIX',
                    'Operating System :: Microsoft :: Windows',
                    'Operating System :: MacOS :: MacOS X',
                    'Topic :: Software Development :: Testing',
                    'Topic :: Software Development :: Libraries',
                    'Programming Language :: Python',
                    'Programming Language :: Python :: 2.7',
                    'Programming Language :: Python :: 3.4',
                    'Programming Language :: Python :: 3.5',
                    'Programming Language :: Python :: 3.6',
                    'Programming Language :: Python :: 3.7'],
    'install_requires': ['selenium', 'six', 'python-dateutil'],
    'package_dir': {'nerodia': 'nerodia'},
    'packages': ['nerodia',
                 'nerodia.elements',
                 'nerodia.locators',
                 'nerodia.locators.button',
                 'nerodia.locators.cell',
                 'nerodia.locators.element',
                 'nerodia.locators.row',
                 'nerodia.locators.text_area',
                 'nerodia.locators.text_field',
                 'nerodia.wait'],
    'package_data': {
        'nerodia': ['js_snippets/*.js']
    },
    'zip_safe': False
}

setup(**setup_args)



from os import path


class JSSnippet(object):
    # private

    def _execute_js(self, function_name, *args):
        filepath = path.abspath(path.join(path.dirname(__file__),
                                          'js_snippets',
                                          '{}.js'.format(function_name)))
        if not path.isfile(filepath):
            from nerodia.exception import Error
            raise Error('Can not excute script as {!r} does not exist'.format(filepath))

        with open(filepath, 'r') as myfile:
            script = 'return ({}).apply(null, arguments)'.format(myfile.read())
            return self.query_scope.execute_script(script, *args)



from copy import copy

from ..element.locator import Locator as ElementLocator


class Locator(ElementLocator):
    # private

    def _using_selenium(self, *args):
        return None  # force using Nerodia

    @property
    def _can_convert_regexp_to_contains(self):
        # regexp conversion won't work with the complex xpath selector
        return False

    def _matches_selector(self, element, selector):
        from ..element.validator import Validator
        if 'value' in selector:
            cpy = copy(selector)
            value = cpy.pop('value', None)

            return super(Locator, self)._matches_selector(element, cpy) and \
                (Validator.match_str_or_regex(value, self._fetch_value(element, 'value')) or
                 Validator.match_str_or_regex(value, self._fetch_value(element, 'text')))
        else:
            return super(Locator, self)._matches_selector(element, selector)



import time
from json import dump, load


class Cookies(object):
    def __init__(self, driver):
        self.driver = driver

    def __iter__(self):
        """
        Yields each cookie in cookie collection

        :rtype: iter

        :Example:

        list(browser.cookies)
        #=> [{name: 'my_session', value: 'BAh7B0kiD3Nlc3Npb25faWQGOgZFRkk', domain: 'mysite.com'}]
        """
        cookies = self.driver.get_cookies()
        for cookie in cookies:
            expire = cookie.get('expires')
            cookie['expires'] = self._to_time(expire) if expire else None
            yield cookie

    @property
    def to_list(self):
        import nerodia
        nerodia.logger.deprecate('Cookies.to_list', 'list(self)')
        return list(self)

    def __getitem__(self, name):
        """
        Returns a cookie by name
        :param name: name of cookie
        :rtype: dict

        :Example:

        browser.cookies['my_session']
        #=> {name: 'my_session', value: 'BAh7B0kiD3Nlc3Npb25faWQGOgZFRkk', domain: 'mysite.com'}
        """
        return next((c for c in self if c.get('name') == name), None)

    def add(self, name, value, **kwargs):
        """
        Adds new cookie

        :Example:

        browser.cookies.add('my_session', 'BAh7B0kiD3Nlc3Npb25faWQGOgZFRkk', domain='mysite.com')
        """
        cookie = {'name': name, 'value': value}
        keys = list(kwargs)
        if 'secure' in keys:
            cookie['secure'] = kwargs.get('secure')
        if 'path' in keys:
            cookie['path'] = kwargs.get('path')
        if 'expires' in keys:
            cookie['expires'] = kwargs.get('expires')
        if 'domain' in keys:
            cookie['domain'] = kwargs.get('domain')
        self.driver.add_cookie(cookie)

    def delete(self, name):
        """ Deletes cookie by given name """
        self.driver.delete_cookie(name)

    def clear(self):
        """ Deletes all cookies """
        self.driver.delete_all_cookies()

    def save(self, file='.cookies'):
        """
        Save cookies to file

        :param file: file path

        :Example:

        browser.cookies.save('.cookies')
        """
        with open(file, 'w') as f:
            dump(list(self), f, indent=4)

    def load(self, file='.cookies'):
        """
        Loads cookies from file

        :param file: file path

        :Example:

        browser.cookies.load('.cookies')
        """
        with open(file, 'r') as f:
            for cookie in load(f):
                self.add(cookie.pop('name'), cookie.pop('value'), **cookie)

    # private

    def _to_time(self, dt):
        return time.strftime('%Y-%m-%d %H:%M:%S %z', time.localtime(time.mktime(dt.timetuple())))



from copy import copy

from ..element.locator import Locator as ElementLocator


class Locator(ElementLocator):
    # private

    def _using_selenium(self, *args):
        return None  # force using Nerodia

    def _matches_selector(self, element, selector):
        selector = copy(selector)

        tag_name = element.tag_name.lower()

        for key in ['text', 'value', 'label']:
            if key in selector:
                correct_key = 'value' if tag_name == 'input' else 'text'
                selector[correct_key] = selector.pop(key)

        return super(Locator, self)._matches_selector(element, selector)



import six

from nerodia.exception import Error
from .input import Input
from ..meta_elements import MetaHTMLElement


@six.add_metaclass(MetaHTMLElement)
class Button(Input):
    """
    Class representing button elements

    This class covers both <button> and <input type="submit|reset|image|button" /> elements
    """
    VALID_TYPES = ['button', 'reset', 'submit', 'image']

    @property
    def text(self):
        """
        Returns the text of the button.

        For input elements, returns the 'value' attribute.
        For button elements, returns the inner text.

        :rtype: str
        """
        tn = self.tag_name

        if tn == 'input':
            return self.value
        elif tn == 'button':
            return super(Button, self).text
        else:
            raise Error("unknown tag name for button: {}".format(tn))



class Error(Exception):
    pass


class UnknownObjectException(Error):
    pass


class ObjectDisabledException(Error):
    pass


class ObjectReadOnlyException(Error):
    pass


class NoValueFoundException(Error):
    pass


class UnknownCellException(Error):
    pass


class NoMatchingWindowFoundException(Error):
    pass


class UnknownFrameException(Error):
    pass


class UnknownRowException(Error):
    pass



import six

from .html_elements import HTMLElement
from ..cell_container import CellContainer
from ..meta_elements import MetaHTMLElement


@six.add_metaclass(MetaHTMLElement)
class TableRow(CellContainer, HTMLElement):
    def __iter__(self):
        """
        Yields each TableCell associated with this row
        :rtype: iter

        :Example:

        row = browser.tr(id='table_row')
        for cell in row:
            print(cell.text)
        """
        for cell in self.cells():
            yield cell

    def __getitem__(self, idx):
        """
        Get the nth cell (<th> or <td>) of this row
        :param idx: index to get
        :rtype: nerodia.elements.cell.Cell
        """
        return self.cell(index=idx)



from ..element.validator import Validator as ElementValidator
from ...elements.button import Button


class Validator(ElementValidator):
    @staticmethod
    def validate(element, selector):
        tag_name = element.tag_name.lower()
        if tag_name not in ['input', 'button']:
            return None
        # TODO - Verify this is desired behavior based on
        # https://bugzilla.mozilla.org/show_bug.cgi?id=1290963
        if tag_name == 'input' and element.get_attribute('type').lower() not in Button.VALID_TYPES:
            return None
        return element



class XpathSupport(object):
    @staticmethod
    def escape(value):
        if "'" in value:
            parts = ["'{}'".format(part) for part in value.split("'", -1)]
            string = ",\"'\",".join(parts)

            return 'concat({})'.format(string)
        else:
            return "'{}'".format(value)

    @staticmethod
    def lower(value):
        return "translate({},'ABCDEFGHIJKLMNOPQRSTUVWXYZ'," \
               "'abcdefghijklmnopqrstuvwxyz')".format(value)



import six

from .element import Element
from ..element_collection import ElementCollection
from ..meta_elements import MetaSVGElement


@six.add_metaclass(MetaSVGElement)
class SVGElement(Element):
    pass


@six.add_metaclass(MetaSVGElement)
class SVGElementCollection(ElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class MPath(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class MPathCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Animation(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class AnimationCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class AnimateTransform(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class AnimateTransformCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class AnimateMotion(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class AnimateMotionCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Set(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class SetCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Animate(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class AnimateCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class View(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class ViewCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Cursor(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class CursorCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Pattern(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class PatternCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Stop(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class StopCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class MeshPatch(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class MeshPatchCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class MeshRow(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class MeshRowCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Gradient(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class GradientCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class MeshGradient(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class MeshGradientCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class RadialGradient(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class RadialGradientCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class LinearGradient(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class LinearGradientCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Marker(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class MarkerCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Symbol(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class SymbolCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Metadata(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class MetadataCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Desc(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class DescCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Graphics(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class GraphicsCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class ForeignObject(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class ForeignObjectCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class TextContent(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class TextContentCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class TextPath(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class TextPathCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class TextPositioning(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class TextPositioningCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class TSpan(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class TSpanCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Switch(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class SwitchCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Use(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class UseCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Defs(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class DefsCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class G(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class GCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class SVG(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class SVGCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Geometry(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class GeometryCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Polygon(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class PolygonCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Polyline(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class PolylineCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Line(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class LineCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Ellipse(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class EllipseCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Circle(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class CircleCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Rect(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class RectCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Path(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class PathCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Hatchpath(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class HatchpathCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Hatch(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class HatchCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Meshpatch(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class MeshpatchCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Meshrow(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class MeshrowCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Solidcolor(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class SolidcolorCollection(SVGElementCollection):
    pass


@six.add_metaclass(MetaSVGElement)
class Mesh(SVGElement):
    pass


@six.add_metaclass(MetaSVGElement)
class MeshCollection(SVGElementCollection):
    pass



import six

from .html_elements import InputCollection
from .input import Input
from ..meta_elements import MetaHTMLElement


@six.add_metaclass(MetaHTMLElement)
class CheckBox(Input):
    @property
    def is_set(self):
        """
        Returns True if the element is checked
        :rtype: bool
        """
        return self._element_call(lambda: self.el.is_selected())

    is_checked = is_set

    def set(self, value=True):
        """
        Sets checkbox to the given value

        :param value: True to check, False to uncheck

        :Example:

        checkbox = browser.checkbox(id='new_user_interests_cars')
        checkbox.is_set        #=> false
        checkbox.set()
        checkbox.is_set         #=> true
        checkbox.set(False)
        checkbox.set            #=> false
        """
        self._assert_enabled() if self.is_set == value else self.click()

    check = set

    def clear(self):
        self.set(value=False)

    uncheck = clear


@six.add_metaclass(MetaHTMLElement)
class CheckBoxCollection(InputCollection):
    pass
